---
title: "            Optimal transport: mature tools and open problems"
author: "Jean Feydy"
institute: ""
topic: ""
theme: "metropolis"
colortheme:
fonttheme:
fontsize: 11pt
linkcolor: black
urlcolor: blue
linkstyle: bold
aspectratio: 169
titlegraphic: 
logo:
date:
section-titles: false
toc: false
---


## My motivation: speeding up core computations for healthcare


**Computational anatomy.** 3D medical scans are orders of magnitude heavier than natural 2D images:

- 100k triangles to represent a brain surface.
- 512x512x512 $\simeq$ 130M voxels for a typical 3D image.


**Public health.** Over the last decade, medical datasets have **blown up**:

- Clinical trials: **1k patients**, controlled environment.
- UK Biobank: **500k people**, curated data.
- French Health Data Hub: **70M people**, full social security data since ~2000.

Medical doctors, pharmacists and governments need scalable methods.


## A field that is moving fast

::: columns

:::: {.column width=70%}


`\begin{center}`{=latex}
`\vspace*{.5cm}`{=latex}
`\contexte{Target.}`{=latex} Scale up models that combine \
medical **expertise** with modern **datasets**.

`\end{center}`{=latex}

\
`\defi{Context.}`{=latex} The advent of **Graphics Processing Units** (GPU): \
 \

- Incredible **value for money**: \
  $\qquad$ 1 000€ $\simeq$ 1 000 cores $\simeq$ $\text{10}^{\text{12}}$ operations/s.
- **Bottleneck**: constraints on **register** usage.

`\vspace*{.5cm}`{=latex}
`\begin{center}`{=latex}

"User-friendly" Python ecosystem, consolidated around \
a **small number of key operations**.

`\end{center}`{=latex}

::::

:::: {.column width=25%}

`\begin{center}`{=latex}

`\vspace*{1cm}`{=latex}

![](images/ampere.jpg){height=6cm}

`\vspace*{.25cm}`{=latex}

**7,000 cores**\
in a single GPU.

`\end{center}`{=latex}

::::

:::


## The KeOps library: efficient support for symbolic matrices


::: columns

:::: {.column width=25%}

`\begin{center}`{=latex}



```{=latex}
\begin{tikzpicture}[scale=.9]
	% Left border:
	\draw[draw=black, fill=red!20] (-.45,0) rectangle ++(.3,3);
	\draw[draw=black, fill=red!80] (-.45,9*.15) rectangle ++(.3,.15);
	\draw[step=.15, color=gray] (-.45,0) grid ++(.3,3);
	\draw[draw=black] (-.45,0) rectangle ++(.3,3);
	%
	% Top border:
	\draw[draw=black, fill=green!20] (0,3.15) rectangle ++(3,.3);
	\draw[draw=black, fill=green!80] (6*.15,3.15) rectangle ++(.15,.3);
	\draw[step=.15, color=gray] (0,3.15) grid ++(3,.3);
	\draw[draw=green!50!black] (0,3.15) rectangle ++(3,.3);
	%
	\draw[draw=black,thick] (0,0) rectangle ++(3,3);
	%
	\draw[->, thick, color=red!80!black] (.5*.15, 9.5*.15) -- (5*.15, 9.5*.15);
	\draw[->, thick, color=green!50!black] (6.5*.15, 19.5*.15) -- (6.5*.15, 11*.15);
	%
	\draw[fill=blue!40] (6*.15,9*.15) rectangle ++(.15,.15);
	\node[inner sep=0pt, anchor = mid, color=blue!50!black] at (1.95,1.4) {\normalsize{F(\,\textcolor{red!80!black}{$\text{x}_\text{i}$}\,,\,\textcolor{green!50!black}{$\text{y}_\text{j}$}\,)}};
\end{tikzpicture}
```

**Symbolic matrix** \
Formula + data
 

`\end{center}`{=latex}
`\vspace*{.5cm}`{=latex}


- Distances `d(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Kernel `k(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Numerous transforms.


::::

:::: {.column width=70%}

`\solution{Solution.}`{=latex} 
KeOps -- `\textcolor{blue}{\texttt{www.kernel-operations.io}}`{=latex}: `\vspace*{.2cm}`{=latex}

- For PyTorch, NumPy, Matlab and R, on **CPU and GPU**.
- **Automatic differentiation**.
- Just-in-time **compilation** of **optimized** C++ schemes, \
  triggered for every new **reduction**: sum, min, etc.

\
If the formula "F" is simple ($\leqslant$ 100 arithmetic operations): \
$\qquad$ "100k $\times$ 100k" computation  $\rightarrow$  10ms -- 100ms, \
$\qquad$     "1M $\times$ 1M"    computation  $\rightarrow\phantom{\text{0m}}$ 1s -- 10s.

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}

Hardware ceiling of $\text{10}^{\text{12}}$ operations/s. \
**$\times \textbf{10}$ to $\times \textbf{100}$ speed-up** vs standard GPU implementations \
for a wide range of problems.

`\end{center}`{=latex}

::::

:::


## A long-term investment in the foundations of our field


Since 2016, I've been working on speeding up:

- Geometric **machine learning**: K-Nearest Neighbors, kernel methods.
- Geometric **statistics**: Gaussian processes, Maximum Mean Discrepancies.
- Geometric **deep learning**: point convolutions, attention layers.

- **Survival** analysis: CoxPH solvers, time-varying features.

- **Optimal transport**: our focus today!


## Today’s talk

1. What is Optimal Transport, and **why does it matter**? \

2. **Computational** advances. \

3. How people use OT **today**. \

4. What about the **future**?


# Optimal transport?

## Optimal transport (OT) generalizes sorting to spaces of dimension D $>$ 1



::: columns

:::: {.column width=70%}

`\vspace*{.3cm}`{=latex}

If $\a{\text{A} = (\text{x}_\text{1}, \dots, \text{x}_\text{N})}$  and  $\b{\text{B} = (\text{y}_\text{1}, \dots, \text{y}_\text{N})}$ \
are two clouds of N points in $\mathbb{R}^\text{D}$,
we define:


```{=latex}
\vspace*{-.3cm}
\begin{equation*}
\OT(\a{\text{A}},\b{\text{B}}) ~=~ \min_{\sigma \in \mathcal{S}_\text{N}}~ \frac{\text{1}}{\text{2N}} \sum_{\text{i\,=1}}^\text{N} \v{\|~\a{\text{x}_i} - \b{\text{y}_{\sigma(i)}}\|^2}
\end{equation*}
\vspace*{.2cm}
```

Generalizes **sorting** to metric spaces. \
**Linear problem** on the permutation matrix $\g{\text{P}}$:

```{=latex}
\begin{equation*}
\OT(\a{\text{A}},\b{\text{B}}) ~=~ \min_{\g{\text{P}} \in \mathbb{R}^{\text{N}\times \text{N}} ~}
\frac{\text{1}}{\text{2N}} \sum_{\text{i, j\,=1}}^{\text{N}} \g{\text{P}_{i,j}} \cdot \v{ \|~ \a{\text{x}_i} - \b{\text{y}_j} \|^2 }~,
\end{equation*}
\vspace*{-.6cm}
\begin{align*}
&\text{s.t.} &
\g{ \text{P}_{i,j}} ~&\geqslant ~\text{0} &
\underbrace{\b{\textstyle\sum_j} \g{ \text{P}_{i,j}} ~=~ \a{\text{1}}}_{\text{Each source point\dots{}}}& &
\underbrace{\vphantom{\textstyle\sum_j}\a{\textstyle\sum_i} \g{ \text{P}_{i,j}} ~=~ \b{\text{1}}~.}_{\text{is transported onto the target.}}
\end{align*}

```

::::

:::: {.column width=25%}

```{=latex}
\begin{center}
	\begin{tikzpicture}[scale=.5,x=.7cm,y=-.7cm]
		%\node[anchor = east] at (0,4) {source};
		\draw[->,>=latex,thick,color=black]  (0,0)--(0,16.5);
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{1}}$}] (x1) at (0,1) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{2}}$}] (x2) at (0,2) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{3}}$}] (x3) at (0,3.5) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{4}}$}] (x4) at (0,5) {};
		\node[circle,fill=red!80!black,inner sep=0pt,minimum size=7pt,label=left:{$\a{\text{x}_\text{5}}$}] (x5) at (0,7.5) {};


		%\node[anchor = east] at (0,0) {cible};
		\draw[->,>=latex,thick,color=black]  (4,0)--(4,16.5);
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{3}}$}] (y1) at (4,5.5) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{5}}$}] (y2) at (4,7) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{2}}$}] (y3) at (4,8.5) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{4}}$}] (y4) at (4,12.5) {};
		\node[circle,fill=blue,inner sep=0pt,minimum size=7pt,label=right:{$\b{\text{y}_\text{1}}$}] (y5) at (4,15) {};

		\node[anchor = north, align=center, color=otplan!70!black] at (2,17) {{assignment}\\ ${\sigma:\zp[i]{1,5}\rightarrow \zp[i]{1,5}}$};

		\draw[color=otplan,very thick] (x1) -- (y1) (x2) -- (y2) (x3) -- (y3) (x4) -- (y4) (x5) -- (y5);

		%\onslide<2,3,4,6->{\draw[color=green!80!black,thick] (x3) -- (y3) (x4) -- (y4);}
		%\onslide<5>{\draw[color=green!80!black] (x4) -- (y3);
		%\draw[color=green!80!black, very thick] (x3) -- (y4);}

	\end{tikzpicture}
	\end{center}
```

::::

:::

## Practical use

Alternatively, we understand OT as:

- Nearest neighbor **projection** + **incompressibility** constraint. \

- Fundamental example of **linear optimization** over the transport plan $\g{ \text{P}_{i,j}}$.

\
This theory induces two main quantities:

- The transport plan $\g{ \text{P}_{i,j}}$ $\simeq$ the optimal mapping $\a{x_i} \mapsto \b{y_{\sigma(i)}}$. \

- The "Wasserstein" distance $\sqrt{\OT(\a{\text{A}},\b{\text{B}})}$.


## OT induces a geometry-aware distance between probability distributions `\cite{peyre}`{=latex}

**Gauss** map    $\mathcal{N} : (m, \sigma) \in \mathbb{R}\times\mathbb{R}_{\geqslant 0}~~~\mapsto~~~\mathcal{N}(m, \sigma) \in \mathbb{P}(\mathbb{R})$.

If the space of **probability distributions** $\mathbb{P}(\mathbb{R})$ is endowed with a given metric, \
what is the "pull-back" geometry on the space of **parameters** $(m, \sigma)$?

::: columns

:::: {.column width=55%}

`\begin{center}`{=latex}

![](images/wasserstein/Gaussian_fisher_rao.png){height=45%}

Fisher-Rao ($\simeq$ relative entropy) on $\mathcal{N}(m, \sigma)$ \
$\rightarrow$ Hyperbolic Poincaré metric on $(m, \sigma)$.

`\end{center}`{=latex}

::::

:::: {.column width=40%}

`\begin{center}`{=latex}

![](images/wasserstein/Gaussian_wasserstein.png){height=45%}

OT on $\mathcal{N}(m, \sigma)$ \
$\rightarrow$ Flat Euclidean metric on $(m, \sigma)$.

`\end{center}`{=latex}

::::

:::



## Geometric solutions to least square problems `\cite{agueh_2011}`{=latex}

```{=latex}
\vspace*{-.5cm}
\begin{align*}
	\text{Barycenter}~~\a{\al^*}
	~=~\arg \min_{\a{\al}} 
	\sum_{i=1}^4 \,\lambda_i\,\text{Loss}(\,\a{\al}\,,\, \b{\be_i}\,)~.
\end{align*}
```

::: columns

:::: {.column width=45%}

`\begin{center}`{=latex}

![](images/wasserstein/linear_barycenters.png){height=60%}

**Euclidean** barycenters. \
$\text{Loss}(\a{\al},\b{\be}) = \|\a{\al}-\b{\be}\|^2_{L^2}$

`\end{center}`{=latex}

::::

:::: {.column width=45%}

`\begin{center}`{=latex}

![](images/wasserstein/wasserstein_barycenters.png){height=60%}

**Wasserstein** barycenters. \
$\text{Loss}(\a{\al},\b{\be}) = \OT(\a{\al},\b{\be})$

`\end{center}`{=latex}

::::

:::




# How should we solve the OT problem?


## A fundamental problem in applied mathematics

Key dates for discrete optimal transport with N points:

- `\cite{kantorovich1942translocation}`{=latex}: `\defi{Dual}`{=latex} problem of Kantorovitch.
- `\cite{kuhn1955hungarian}`{=latex}: `\defi{Hungarian}`{=latex} methods in $O(\text{N}^3)$.
- `\cite{bertsekas79auction}`{=latex}: **Auction** algorithm in $O(\text{N}^2)$.
- `\cite{kosowsky1994invisible}`{=latex}: **SoftAssign** = Sinkhorn + simulated annealing, in $O(\text{N}^2)$.
- `\cite{gold1998new,TPSRPM}`{=latex}: **Robust Point Matching** = Sinkhorn as a loss.
- `\cite{CuturiSinkhorn}`{=latex}: Start of the `\contexte{GPU era.}`{=latex}
- `\cite{merigot2011multiscale,levy2015numerical,schmitzer2019stabilized}`{=latex}: **multi-scale** solvers in $O(\text{N}\log \text{N})$. \
  $~$
- `\solution{Solution,}`{=latex}today: **Multiscale Sinkhorn algorithm, on the GPU**.


`\begin{center}`{=latex}
$\Longrightarrow$ Generalized `\solution{QuickSort}`{=latex}algorithm.
`\end{center}`{=latex}


## Scaling up optimal transport to anatomical data

```{=latex}

\begin{center} 
	Progresses of the last decade add up to a 
	$\times \textbf{100}$ - $\times \textbf{1000}$ acceleration:
	
	Sinkhorn GPU
	$\xrightarrow{\times \text{10}}$
	+ KeOps
	$\xrightarrow{\times \text{10}}$
	+ Annealing
	$\xrightarrow{\times \text{10}}$
	+ Multi-scale
	
	
	With a precision of 1\%, on a gaming GPU:\\~
	\end{center}

	\vspace*{-.25cm}
	\begin{tabular}{cccc}
	\begin{minipage}{.2\linewidth}
	\begin{center}
	~\\
	\texttt{pip install geomloss}\\
	+\\
	modern GPU\\ (1 000 €)
	\end{center}
	\end{minipage}
	&
	\begin{minipage}{.1\linewidth}
	~\\
	$\Longrightarrow$
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/lowres.png}
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/highres.png}
	\end{minipage}
	\\
	&&
	10k points in 30-50ms
	&
	\textbf{100k points in 100-200ms}
	\end{tabular}
```


# How do people use OT in 2022?

## 1. Physics and simulation of Partial Differential Equations

Since the 1990s, OT is an essential tool to deal with flows:

- Fundamental models have an **appealing form** when seen through the OT lense:\
  the incompressible **Euler flow** is a **geodesic** trajectory,\
  **heat** diffusion is a gradient **descent**...\

- This framework allows mathematicians to design and study new models **effectively**. \

- **Implementations** in 2D and 3D are now becoming mature. \

- Lots of cool simulations of **crowds**, **water** or the **early universe**!

\
**Pointers:** MoKaPlan Inria team, Bruno Lévy, Quentin Mérigot, \
Filippo Santambrogio, Yann Brenier, Felix Otto…


## 2. An intriguing tool in machine learning

OT **lifts to probability distributions** the geometry of the sample space $\|\a{x_i}-\b{y_j}\|$.

This is relevant at the intersection between geometry and statistics in order to:

- Design **2-sample tests** : do these two samples come from the same distribution?
- Quantify the **discrepancy** between a synthetic sample and the data distribution.
- Study the convergence of **particle-based optimization** schemes, \
  from simple neural networks to MCMC samplers.

\
**Pointers:** Python Optimal Transport (Flamary, Courty et al.), \
Computational Optimal Transport (Peyré and Cuturi), \
Jonathan Weed, Justin Solomon, Philippe Rigollet, Lenaïc Chizat, Anna Korba…


## 3. A typical example in shape analysis: lung registration "Exhale -- Inhale"

`\begin{center}`{=latex}

![](images/applications/poumons.png){height=85%}

`\end{center}\vspace*{-.3cm}`{=latex}

**Complex** deformations, high **resolution** (50k--300k points), high **accuracy** ($<$ 1mm).



## State-of-the-art networks -- and their limitations



::: columns

:::: {.column width=45%}

`\begin{center}`{=latex}
`\vspace*{1cm}`{=latex}

![](images/architectures/deep_reg_module.pdf){width=\textwidth}

`\vspace*{.25cm}`{=latex}
**Multi-scale** convolutional \
point neural network.

`\end{center}`{=latex}


::::

:::: {.column width=50%}

`\vspace*{.6cm}`{=latex}
Point neural nets, **in practice**:

- Compute **descriptors** at all scales.
- **Match** them using geometric layers.
- Train on **synthetic** deformations.

\
Strengths and weaknesses:

- Good at **pairing** branches.
- Hard to train to high **accuracy**.

\
$\Longrightarrow$    **Complementary** to OT.

::::


:::




## Three-steps registration with Robust OT (RobOT)

`\begin{center}\vspace*{.2cm}`{=latex}

![](images/drobot/drobot_summary.pdf){width=80%}

`\end{center}\vspace*{-.5cm}`{=latex}

 \
This **pragmatic** method:

- Is **easy to train** on synthetic data.
- Scales up to high-resolution: 100k points in 1s.
- Excellent results: **KITTI** (outdoors scans) and **DirLab** (lungs).

`\begin{center}`{=latex}
***Accurate*** *point cloud registration with **robust** optimal transport*,\
Shen, Feydy et al., NeurIPS 2021.
`\end{center}`{=latex}

## Three-steps registration


`\begin{center}`{=latex}

![](images/applications/lung_registration/full.png){height=85%}

`\end{center}\vspace*{-.3cm}`{=latex}


# What about the future?

## Some open problems

1. Can we generalize standard ML algorithms from **vector spaces**\
   to a (non-linear) space of **probability** distributions? \
    

2. What about distances on **graphs**? \
   What about **non-convex** costs, e.g. $\sqrt{\|\a{x_i}-\b{y_j}\|}$? \
    

3. Can we enforce some **spatial regularity** while keeping super-fast solvers? \
    

4. OT as a source of inspiration in **high-dimension**:\
   can we design robust geometric distances between distributions?

## My job: pave the way for a new generation of researchers

1. **Secure** a permanent position. \
  $\rightarrow$ Inria researcher since Dec. 2021. \

2. Shore up the **GPU foundations** of the field. \
  $\rightarrow$ KeOps v2.0 released in March 2022, now seamless to install.\

3. **Re-write GeomLoss** with a better interface and full support for 2D/3D images. \
  $\rightarrow$ WIP with the Python Optimal Transport devs, first release very soon. \

4. Maintain an **open benchmarking platform** for the community, \
  following the example of `\url{www.ann-benchmarks.com}`{=latex} for nearest neighbor search. \
  $\rightarrow$ WIP, release this Fall.


# Conclusion



## Genuine team work

```{=latex}
\vspace*{1cm}

\centerline{
\begin{tabular}{cccc}
\includegraphics[height=2cm]{images/sourires/medaillons/Alain_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Thibault_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/FX_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Gabriel_2.png}
\\
Alain Trouvé
&
\textcolor{red!65!black}{Thibault Séjourné}
&
\textcolor{red!65!black}{F.-X. Vialard}
&
\textcolor{red!65!black}{Gabriel Peyré}
\\[.5cm]
\includegraphics[height=2cm]{images/sourires/medaillons/Benjamin_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Joan_2.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Marc.png}
&
\includegraphics[height=2cm]{images/sourires/medaillons/Zhengyang.png}\\
\textcolor{blue!100!black}{Benjamin Charlier}
&
\textcolor{blue!100!black}{Joan Glaunès}
&
\textcolor{green!60!black}{Marc Niethammer}
&
\textcolor{green!60!black}{Shen Zhengyang}\\[.5cm]
\end{tabular}
}

```

## We're hiring!

Job opportunities at Inria Paris:

- Research **engineer**: Deformetrica/scikit-shapes, Sep. 2022 -- Feb. 2024.

- Paid M2 **internship**: social security data, Apr. 2023 -- Aug. 2023.

- **PhD thesis**: geometric data analysis, Sep. 2023 -- Aug. 2026.


## Key points

- `\solution{Symbolic}`{=latex}matrices are to **geometric** ML what \
       **sparse** matrices are to **graph** processing: \
  $\longrightarrow~~$ KeOps: **x30 speed-up** vs. PyTorch, TF et JAX. \
  $\longrightarrow~~$ Useful in a wide range of settings. \
  \

- Optimal Transport = `\solution{generalized sorting}`{=latex} = **incompressibility** prior: \
  $\longrightarrow~~$ Super-fast solvers on simple domains (especially 2D/3D spaces).\
  $\longrightarrow~~$ Fundamental tool at the intersection of geometry and statistics. \
  \
  
- GPUs are more **versatile** than you think. \
  $\longrightarrow~~$ Ongoing work to provide **fast GPU backends** to researchers, \
  $\phantom{\longrightarrow}~~~$ going beyond what Google and Facebook are ready to pay for. \



## Documentation and tutorials are available online

```{=latex}
\begin{center}
$\Longrightarrow$\qquad \url{www.kernel-operations.io}\qquad $\Longleftarrow$
\vspace*{.15cm}

\includegraphics[width=.5\linewidth,frame]{images/geometric_data_analysis.png}

%\vspace*{.15cm}

\url{www.jeanfeydy.com/geometric_data_analysis.pdf}
\end{center}

```



# References


